#ifndef ZP_ASSEMBLY_GENERATOR_H_
#define ZP_ASSEMBLY_GENERATOR_H_

#include "Optimizer.h"
#include <iostream>
#include <string>

namespace zp
{
    class IRVisitor;

    struct LabelNames;

    class AssemblyGenerator;
}


class zp::IRVisitor
{
public:
    virtual void visitBlockInstruction(const BlockInstruction&) = 0;
    virtual void visitUnionBlockInstruction(const UnionBlockInstruction&) = 0;
    virtual void visitIOInstruction(const IOInstruction&) = 0;
    virtual void visitSimpleBlockInstruction(const SimpleBlockInstruction&) = 0;
    virtual void visitLinearLoopInstruction(const LinearLoopInstruction&) = 0;
    virtual void visitLoopInstruction(const LoopInstruction&) = 0;
};


struct zp::LabelNames
{
    static const char* main;
    static const char* getchar;
    static const char* putchar;
    static const char* calloc;
    static const char* free;
};


class zp::AssemblyGenerator :
    public IRVisitor
{
    std::ostream& out;

    int labelcount = 0;
    static const char* dataRegister;
    static const char* pointerRegister;
    static const char* argumentRegister1;
    static const char* argumentRegister2;
public:

    AssemblyGenerator(std::ostream& out);

    void finish(void);

    std::string generateLabel(void);
    std::string generateOffset(const std::string& offset, const std::string& base);
    std::string ptr(void);

    template<typename... Args>
    void write(Args... args)
    {
        (out << ... << args) << std::endl;
    }

    virtual void visitBlockInstruction(const BlockInstruction&);
    virtual void visitUnionBlockInstruction(const UnionBlockInstruction&);
    virtual void visitIOInstruction(const IOInstruction&);
    virtual void visitSimpleBlockInstruction(const SimpleBlockInstruction&);
    virtual void visitLinearLoopInstruction(const LinearLoopInstruction&);
    virtual void visitLoopInstruction(const LoopInstruction&);

    void compileMult(Index index, Cell factor);
    int trailingZeroes(unsigned long long x);
};




#endif // ZP_ASSEMBLY_GENERATOR_H_
