#include "Optimizer.h"
#include "Parser.h"
#include "AssemblyGenerator.h"
#include <iostream>
#include <vector>

using namespace zp;
using namespace std;

unique_ptr<BlockInstruction> AstVisitor::visit(Block& block)
{
    return block.accept(*this);
}


void UnionBlockInstruction::accept(IRVisitor& v)
{
    v.visitUnionBlockInstruction(*this);
}


void IOInstruction::accept(IRVisitor& v)
{
    v.visitIOInstruction(*this);
}


void SimpleBlockInstruction::accept(IRVisitor& v)
{
    v.visitSimpleBlockInstruction(*this);
}


void LinearLoopInstruction::accept(IRVisitor& v)
{
    v.visitLinearLoopInstruction(*this);
}


void LoopInstruction::accept(IRVisitor& v)
{
    v.visitLoopInstruction(*this);
}


unique_ptr<BlockInstruction> Optimizer::visitInstructionBlock(InstructionBlock& ib)
{
    unique_ptr<SimpleBlockInstruction> sbi = make_unique<SimpleBlockInstruction>();
    for (auto instr : ib.instructions) {
        switch(instr) {
            case Instruction::LEFT:
                sbi->pointerMoved--;
                break;
            case Instruction::RIGHT:
                sbi->pointerMoved++;
                break;
            case Instruction::PLUS:
                if (sbi->added.find(sbi->pointerMoved) == sbi->added.end())
                    sbi->added[sbi->pointerMoved] = 0;
                sbi->added[sbi->pointerMoved]++;
                break;
            case Instruction::MINUS:
                if (sbi->added.find(sbi->pointerMoved) == sbi->added.end())
                    sbi->added[sbi->pointerMoved] = 0;
                sbi->added[sbi->pointerMoved]--;
                break;
            default:
                break;
        }
    }
    /*for (const auto [index, value] : sbi->added) {
        std::cout << "we have: " << index << ", " << int(value) << std::endl;
    }*/
    return sbi;
}


unique_ptr<BlockInstruction> Optimizer::visitIOBlock(IOBlock& io)
{
    return make_unique<IOInstruction>(io.ioInstruction == Instruction::COMMA);
}


unique_ptr<BlockInstruction> Optimizer::visitUnionBlock(UnionBlock& ub)
{
    auto ret = make_unique<UnionBlockInstruction>();
    ret->content.reserve(ub.instructions.size());
    for (auto& block : ub.instructions) {
        ret->content.push_back(visit(*block.get()));
    }
    return ret;
}


unique_ptr<BlockInstruction> Optimizer::visitLoop(Loop& loop)
{
    auto ret = make_unique<LoopInstruction>();
    ret->content.reserve(loop.instructions.size());
    bool noSubLoops = true;
    Index pointerMoved = 0;
    for (auto& block : loop.instructions) {
        auto v = visit(*block.get());
        if (auto sbi = dynamic_cast<SimpleBlockInstruction*>(v.get())) {
            pointerMoved += sbi->pointerMoved;
        }
        else {
            noSubLoops = false;
        }
        ret->content.push_back(std::move(v));
    }

    if (noSubLoops && pointerMoved == 0 && ret->content.size() == 1) {
        auto sbi = dynamic_cast<SimpleBlockInstruction*>(ret->content[0].get());
        auto linear = make_unique<LinearLoopInstruction>();
        for (const auto [index, value] : sbi->added) {
            if (index == 0)
                linear->increment = value;
            else
                linear->factors[index] = value;
        }
        if (linear->increment == 255) {
            return linear;
        }
    }
    return ret;
}


